use std::collections::HashSet;

use diesel::prelude::*;
use serde::{Deserialize, Serialize};

use crate::db::{schema::*, DbPool, DbResult};

#[derive(Serialize, Deserialize, Clone, Queryable, Selectable)]
#[diesel(table_name = subscriptions)]
#[diesel(check_for_backend(diesel::sqlite::Sqlite))]
pub struct Subscription {
    pub id: i64,
    pub device_id: i64,
    pub url: String,
    pub time_changed: i64,
    pub deleted: bool,
}

#[derive(Deserialize, Insertable)]
#[diesel(table_name = subscriptions)]
#[diesel(check_for_backend(diesel::sqlite::Sqlite))]
pub struct NewSubscription {
    pub device_id: i64,
    pub url: String,
    pub time_changed: i64,
    pub deleted: bool,
}

impl Subscription {
    pub fn for_device(pool: &DbPool, device_id: i64) -> DbResult<Vec<String>> {
        Ok(subscriptions::table
            .select(subscriptions::url)
            .filter(subscriptions::device_id.eq(device_id))
            .get_results(&mut pool.get()?)?)
    }

    pub fn for_user(pool: &DbPool, user_id: i64) -> DbResult<Vec<String>> {
        Ok(subscriptions::table
            .inner_join(devices::table)
            .filter(devices::user_id.eq(user_id))
            .select(subscriptions::url)
            .distinct()
            .get_results(&mut pool.get()?)?)
    }

    pub fn set_for_device(
        pool: &DbPool,
        device_id: i64,
        urls: Vec<String>,
        timestamp: i64,
    ) -> DbResult<()> {
        pool.get()?.transaction(|conn| {
            // https://github.com/diesel-rs/diesel/discussions/2826
            // SQLite doesn't support default on conflict set values, so we can't handle this using
            // on conflict. Therefore, we instead calculate which URLs should be inserted and which
            // updated, so we avoid conflicts.
            let urls: HashSet<String> = urls.into_iter().collect();
            let urls_in_db: HashSet<String> = subscriptions::table
                .select(subscriptions::url)
                .filter(subscriptions::device_id.eq(device_id))
                .get_results(&mut pool.get()?)?
                .into_iter()
                .collect();

            // URLs originally in the database that are no longer in the list
            let urls_to_delete = urls_in_db.difference(&urls);

            // URLs not in the database that are in the new list
            let urls_to_insert = urls.difference(&urls_in_db);

            // URLs that are in both the database and the new list. For these, those marked as
            // "deleted" in the database are updated so they're no longer deleted, with their
            // timestamp updated.
            let urls_to_update = urls.intersection(&urls_in_db);

            // Mark the URLs to delete as properly deleted
            diesel::update(
                subscriptions::table.filter(
                    subscriptions::device_id
                        .eq(device_id)
                        .and(subscriptions::url.eq_any(urls_to_delete)),
                ),
            )
            .set((
                subscriptions::deleted.eq(true),
                subscriptions::time_changed.eq(timestamp),
            ))
            .execute(conn)?;

            // Update the existing deleted URLs that are reinserted as no longer deleted
            diesel::update(
                subscriptions::table.filter(
                    subscriptions::device_id
                        .eq(device_id)
                        .and(subscriptions::url.eq_any(urls_to_update))
                        .and(subscriptions::deleted.eq(true)),
                ),
            )
            .set((
                subscriptions::deleted.eq(false),
                subscriptions::time_changed.eq(timestamp),
            ))
            .execute(conn)?;

            // Insert the new values into the database
            diesel::insert_into(subscriptions::table)
                .values(
                    urls_to_insert
                        .into_iter()
                        .map(|url| NewSubscription {
                            device_id,
                            url: url.to_string(),
                            deleted: false,
                            time_changed: timestamp,
                        })
                        .collect::<Vec<_>>(),
                )
                .execute(conn)?;

            Ok(())
        })
    }

    pub fn update_for_device(
        pool: &DbPool,
        device_id: i64,
        added: Vec<String>,
        removed: Vec<String>,
        timestamp: i64,
    ) -> DbResult<()> {
        // TODO URLs that are in both the added and removed lists will currently get "re-added",
        // meaning their change timestamp will be updated even though they haven't really changed.
        let added: HashSet<_> = added.into_iter().collect();
        let removed: HashSet<_> = removed.into_iter().collect();

        pool.get()?.transaction(|conn| {
            let urls_in_db: HashSet<String> = subscriptions::table
                .select(subscriptions::url)
                .filter(subscriptions::device_id.eq(device_id))
                .get_results(&mut pool.get()?)?
                .into_iter()
                .collect();

            // Subscriptions to remove are those that were already in the database and are now part
            // of the removed list. Subscriptions that were never added in the first place don't
            // need to be marked as deleted. We also only update those that aren't already marked
            // as deleted.
            let urls_to_delete = removed.intersection(&urls_in_db);

            diesel::update(
                subscriptions::table.filter(
                    subscriptions::device_id
                        .eq(device_id)
                        .and(subscriptions::url.eq_any(urls_to_delete))
                        .and(subscriptions::deleted.eq(false)),
                ),
            )
            .set((
                subscriptions::deleted.eq(true),
                subscriptions::time_changed.eq(timestamp),
            ))
            .execute(conn)?;

            // Subscriptions to update are those that are already in the database, but are also in
            // the added list. Only those who were originally marked as deleted get updated.
            let urls_to_update = added.intersection(&urls_in_db);

            diesel::update(
                subscriptions::table.filter(
                    subscriptions::device_id
                        .eq(device_id)
                        .and(subscriptions::url.eq_any(urls_to_update))
                        .and(subscriptions::deleted.eq(true)),
                ),
            )
            .set((
                subscriptions::deleted.eq(false),
                subscriptions::time_changed.eq(timestamp),
            ))
            .execute(conn)?;

            // Subscriptions to insert are those that aren't in the database and are part of the
            // added list
            let urls_to_insert = added.difference(&urls_in_db);

            diesel::insert_into(subscriptions::table)
                .values(
                    urls_to_insert
                        .into_iter()
                        .map(|url| NewSubscription {
                            device_id,
                            url: url.to_string(),
                            deleted: false,
                            time_changed: timestamp,
                        })
                        .collect::<Vec<_>>(),
                )
                .execute(conn)?;

            Ok(())
        })
    }

    pub fn updated_since_for_device(
        pool: &DbPool,
        device_id: i64,
        timestamp: i64,
    ) -> DbResult<Vec<Self>> {
        Ok(subscriptions::table
            .select(Self::as_select())
            .filter(
                subscriptions::device_id
                    .eq(device_id)
                    .and(subscriptions::time_changed.ge(timestamp)),
            )
            .get_results(&mut pool.get()?)?)
    }
}
