use axum::{
    extract::{Path, State},
    routing::post,
    Router,
};
use axum_extra::{
    extract::{
        cookie::{Cookie, Expiration},
        CookieJar,
    },
    headers::{authorization::Basic, Authorization},
    TypedHeader,
};

use crate::{
    db::{Session, User},
    server::{
        error::{AppError, AppResult},
        gpodder::SESSION_ID_COOKIE,
        Context,
    },
};

pub fn router() -> Router<Context> {
    Router::new()
        .route("/{username}/login.json", post(post_login))
        .route("/{username}/logout.json", post(post_logout))
}

async fn post_login(
    State(ctx): State<Context>,
    Path(username): Path<String>,
    jar: CookieJar,
    TypedHeader(auth): TypedHeader<Authorization<Basic>>,
) -> AppResult<CookieJar> {
    // These should be the same according to the spec
    if username != auth.username() {
        return Err(AppError::BadRequest);
    }

    let session = tokio::task::spawn_blocking(move || {
        let user = User::by_username(&ctx.pool, auth.username())?.ok_or(AppError::NotFound)?;

        if user.verify_password(auth.password()) {
            Ok(Session::new_for_user(&ctx.pool, user.id)?)
        } else {
            Err(AppError::Unauthorized)
        }
    })
    .await
    .unwrap()?;

    Ok(jar.add(
        Cookie::build((SESSION_ID_COOKIE, session.id.to_string())).expires(Expiration::Session),
    ))
}

async fn post_logout(
    State(ctx): State<Context>,
    Path(username): Path<String>,
    jar: CookieJar,
) -> AppResult<CookieJar> {
    if let Some(session_id) = jar.get(SESSION_ID_COOKIE) {
        let session_id: i64 = session_id
            .value()
            .parse()
            .map_err(|_| AppError::BadRequest)?;

        tokio::task::spawn_blocking(move || {
            if let Some(session) = Session::by_id(&ctx.pool, session_id)? {
                let user = session.user(&ctx.pool)?.ok_or(AppError::NotFound)?;

                // The requested user to logout should be the same as the one linked to the session
                // ID
                if username == user.username {
                    Ok(session.remove(&ctx.pool)?)
                } else {
                    Err(AppError::BadRequest)
                }
            } else {
                Ok(false)
            }
        })
        .await
        .unwrap()?;

        Ok(jar.remove(SESSION_ID_COOKIE))
    } else {
        Ok(jar)
    }
}
