use axum::{
    extract::{Path, Query, State},
    middleware,
    routing::post,
    Extension, Json, Router,
};
use serde::Deserialize;

use crate::{
    db,
    server::{
        error::{AppError, AppResult},
        gpodder::{
            auth_middleware,
            format::{Format, StringWithFormat},
            models::{
                DeviceType, SubscriptionChangeResponse, SubscriptionDelta,
                SubscriptionDeltaResponse,
            },
        },
        Context,
    },
};

pub fn router(ctx: Context) -> Router<Context> {
    Router::new()
        .route(
            "/{username}/{id}",
            post(post_subscription_changes).get(get_subscription_changes),
        )
        .layer(middleware::from_fn_with_state(ctx.clone(), auth_middleware))
}

pub async fn post_subscription_changes(
    State(ctx): State<Context>,
    Path((username, id)): Path<(String, StringWithFormat)>,
    Extension(user): Extension<db::User>,
    Json(delta): Json<SubscriptionDelta>,
) -> AppResult<Json<SubscriptionChangeResponse>> {
    if id.format != Format::Json {
        return Err(AppError::NotFound);
    }

    if username != user.username {
        return Err(AppError::BadRequest);
    }

    let timestamp = chrono::Utc::now().timestamp_millis();

    tokio::task::spawn_blocking(move || {
        let device = if let Some(device) = db::Device::by_device_id(&ctx.pool, user.id, &id)? {
            device
        } else {
            db::NewDevice::new(
                user.id,
                id.to_string(),
                String::new(),
                DeviceType::Other.into(),
            )
            .insert(&ctx.pool)?
        };

        db::Subscription::update_for_device(
            &ctx.pool,
            device.id,
            delta.add,
            delta.remove,
            timestamp,
        )
    })
    .await
    .unwrap()?;

    Ok(Json(SubscriptionChangeResponse {
        timestamp: timestamp + 1,
        // TODO implement URL sanitization
        update_urls: vec![],
    }))
}

#[derive(Deserialize)]
pub struct SinceQuery {
    #[serde(default)]
    since: i64,
}

pub async fn get_subscription_changes(
    State(ctx): State<Context>,
    Path((username, id)): Path<(String, StringWithFormat)>,
    Extension(user): Extension<db::User>,
    Query(query): Query<SinceQuery>,
) -> AppResult<Json<SubscriptionDeltaResponse>> {
    if id.format != Format::Json {
        return Err(AppError::NotFound);
    }

    if username != user.username {
        return Err(AppError::BadRequest);
    }

    let subscriptions = tokio::task::spawn_blocking(move || {
        let device =
            db::Device::by_device_id(&ctx.pool, user.id, &id)?.ok_or(AppError::NotFound)?;

        Ok::<_, AppError>(db::Subscription::updated_since_for_device(
            &ctx.pool,
            device.id,
            query.since,
        )?)
    })
    .await
    .unwrap()?;

    let mut delta = SubscriptionDeltaResponse::default();
    delta.timestamp = query.since;

    for sub in subscriptions.into_iter() {
        if sub.deleted {
            delta.remove.push(sub.url);
        } else {
            delta.add.push(sub.url);
        }

        delta.timestamp = delta.timestamp.max(sub.time_changed);
    }

    // Timestamp should reflect the events *after* the last seen change
    delta.timestamp += 1;

    Ok(Json(delta))
}
