use axum::{
    extract::{Path, State},
    middleware,
    routing::get,
    Extension, Form, Json, Router,
};

use crate::{
    db,
    server::{
        error::{AppError, AppResult},
        gpodder::{auth_middleware, format::StringWithFormat, models::DeviceType},
        Context,
    },
};

pub fn router(ctx: Context) -> Router<Context> {
    Router::new()
        .route(
            "/{username}/{id}",
            get(get_device_subscriptions).put(put_device_subscriptions),
        )
        .route("/{username}", get(get_user_subscriptions))
        .layer(middleware::from_fn_with_state(ctx.clone(), auth_middleware))
}

pub async fn get_device_subscriptions(
    State(ctx): State<Context>,
    Path((username, id)): Path<(String, StringWithFormat)>,
    Extension(user): Extension<db::User>,
) -> AppResult<Json<Vec<String>>> {
    if username != user.username {
        return Err(AppError::BadRequest);
    }

    let subscriptions = tokio::task::spawn_blocking(move || {
        let device =
            db::Device::by_device_id(&ctx.pool, user.id, &id)?.ok_or(AppError::NotFound)?;

        Ok::<_, AppError>(db::Subscription::for_device(&ctx.pool, device.id)?)
    })
    .await
    .unwrap()?;

    Ok(Json(subscriptions))
}

pub async fn get_user_subscriptions(
    State(ctx): State<Context>,
    Path(username): Path<StringWithFormat>,
    Extension(user): Extension<db::User>,
) -> AppResult<Json<Vec<String>>> {
    if *username != user.username {
        return Err(AppError::BadRequest);
    }

    let subscriptions =
        tokio::task::spawn_blocking(move || db::Subscription::for_user(&ctx.pool, user.id))
            .await
            .unwrap()?;

    Ok(Json(subscriptions))
}

pub async fn put_device_subscriptions(
    State(ctx): State<Context>,
    Path((username, id)): Path<(String, StringWithFormat)>,
    Extension(user): Extension<db::User>,
    Json(urls): Json<Vec<String>>,
) -> AppResult<()> {
    if *username != user.username {
        return Err(AppError::BadRequest);
    }

    tokio::task::spawn_blocking(move || {
        let device = if let Some(device) = db::Device::by_device_id(&ctx.pool, user.id, &id)? {
            device
        } else {
            db::NewDevice::new(
                user.id,
                id.to_string(),
                String::new(),
                DeviceType::Other.into(),
            )
            .insert(&ctx.pool)?
        };

        Ok::<_, AppError>(db::Subscription::set_for_device(
            &ctx.pool,
            device.id,
            urls,
            chrono::Utc::now().timestamp(),
        )?)
    })
    .await
    .unwrap()?;

    Ok(())
}
